////////////////////////////////////////////////////////////////////////////////
// 
//  Copyright (c) 2010 Renaun Erickson <renaun.com>
// 
//  Permission is hereby granted, free of charge, to any person
//  obtaining a copy of this software and associated documentation
//  files (the "Software"), to deal in the Software without
//  restriction, including without limitation the rights to use,
//  copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following
//  conditions:
// 
//  The above copyright notice and this permission notice shall be
//  included in all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
//  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
//  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
//  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
//  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//  OTHER DEALINGS IN THE SOFTWARE.
// 
////////////////////////////////////////////////////////////////////////////////
/**
 Flex SDK Coding Conventions Used:
 http://opensource.adobe.com/wiki/display/flexsdk/Coding+Conventions
 **/
package com.renaun.physicalscreen
{
	import flash.display.DisplayObject;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.system.Capabilities;
	import flash.utils.Dictionary;
	
	import mx.core.IVisualElement;
	import mx.core.IVisualElementContainer;
	import mx.core.UIComponent;
	import mx.core.mx_internal;
	import mx.events.FlexEvent;
	
	import spark.components.Group;
	import spark.components.SkinnableContainer;
	import spark.events.ElementExistenceEvent;
	
	public class PhysicalScreenUtil
	{
		//--------------------------------------------------------------------------
		//
		//  Static
		//
		//--------------------------------------------------------------------------
		
		private static var styles:Array = ["left", "right", "top", "bottom", 
					"paddingLeft", "paddingRight", "paddingTop", "paddingBottom"];
		private static var ratio:Number = -1;
		private static var originalStyleValues:Dictionary = new Dictionary(true);
		private static var scrollZoomLayoutMouseTrack:Dictionary = new Dictionary(true);
		private static var lastRatio:Number = 1;
		private static var lastFilterStyles:Array = styles;
		
		public static var ppiScaleFactor:Number = 1;//50/23;
		public static var ppiBaseline:Number = 254;
		
		public static function getPPIRatio():Number
		{
			if (ratio < 0)
			{
				ratio = Math.ceil((Capabilities.screenDPI / ppiBaseline) * 100)/100;
			}
			if (Capabilities.screenDPI == 72)
				return 0.5 * ppiScaleFactor;
			else
				return ratio * ppiScaleFactor;
		}
		
		public static function relayout(component:UIComponent, newRatio:Number, 
														filterStyles:Array = null):void
		{
			if (!filterStyles)
				filterStyles = styles;
			lastRatio = newRatio;
			lastFilterStyles = filterStyles;
			//component.addEventListener(FlexEvent.ADD, addedHandler2, false, 0, true);
			if (component.parent)
			{
				addedDLHandler(null, component);
			}
			else
			{
				component.addEventListener(ElementExistenceEvent.ELEMENT_ADD, addedHandler, false, 0, true);
				component.addEventListener(Event.ADDED_TO_STAGE, addedDLHandler, false, 0, true);
			}
		}
		
		public static function addedDLHandler(event:Event, nonEvent:UIComponent = null):void
		{
			var component:UIComponent = nonEvent;
			if (event)
			{
				//trace(event.type + ": " + event.target + " - " + event.currentTarget);
				component = event.target as UIComponent;
			}
			var len:int = (component as IVisualElementContainer).numElements;
			var i:int = 0;
			var element:UIComponent;
			var newRatio:Number = lastRatio;
			var filterStyles:Array = lastFilterStyles;
			for (i = 0;i < len; i++)
			{
				element = (component as IVisualElementContainer).getElementAt(i) as UIComponent;
				resizeComponent(element, newRatio, filterStyles);
			}
			
			len = component.numChildren;
			for (i = 0;i < len; i++)
			{
				element = component.getChildAt(i) as UIComponent;
				if (component.hasOwnProperty("skin") && element != component["skin"])
					resizeComponent(element, newRatio, filterStyles);
			}
		}
		
		protected static function creationComplete(event:Event):void
		{
			//trace(event.type + ": " + event.target + " - " + event.currentTarget);
			var comp:UIComponent = (event.target as UIComponent);
			if (!comp)
				return;
			
			//trace(event.type + ": " + event.target + " - " + event.currentTarget);
			if (event.target.stage)
			{
				//trace("WIDTH1: " + comp.getLayoutBoundsWidth() + " - " + event.target.stage.stageWidth);
				PhysicalScreenUtil.resizeComponent(comp, lastRatio, lastFilterStyles);
			}
			if (event.target.stage && (comp.getLayoutBoundsWidth() > event.target.stage.stageWidth
				|| comp.getLayoutBoundsHeight() > event.target.stage.stageHeight))
			{
				var object:Object = {scrollX: comp.getLayoutBoundsWidth() - event.target.stage.stageWidth,
					scrollY: comp.getLayoutBoundsHeight() - event.target.stage.stageHeight,
					point: null
				};
				if (scrollZoomLayoutMouseTrack[comp.name] && object.scrollX == scrollZoomLayoutMouseTrack[comp.name].scrollX)
				{
					comp.removeEventListener(FlexEvent.UPDATE_COMPLETE, creationComplete);
					return;
				}
				//trace("WIDTH: " + comp.getLayoutBoundsWidth() + " - " + event.target.stage.stageWidth);
				scrollZoomLayoutMouseTrack[comp.name] = object;
				if (!comp.hasEventListener(MouseEvent.MOUSE_DOWN))
				{
					comp.addEventListener(MouseEvent.MOUSE_DOWN, mouseEventHandler, false, 0, true);
					comp.addEventListener(MouseEvent.MOUSE_MOVE, mouseEventHandler, false, 0, true);
					comp.addEventListener(MouseEvent.MOUSE_UP, mouseEventHandler, false, 0, true);
				}
			}
		}
		
		protected static function addedHandler(event:ElementExistenceEvent):void
		{
			//trace(event.target + " - " + event.currentTarget + " - " + event.element);
			//trace("width1: " + event.element.getLayoutBoundsWidth());
			var addedElement:UIComponent = event.element as UIComponent;
			if (addedElement)
			{
				if (!scrollZoomLayoutMouseTrack[addedElement.name])
					addedElement.addEventListener(FlexEvent.UPDATE_COMPLETE, creationComplete, false, 0, true);
			}
			//if (event.target.stage)
			//	trace("width2: " + addedElement.getLayoutBoundsWidth() + " - " + event.target.stage.stageWidth);
		}
		
		protected static function mouseEventHandler(event:MouseEvent):void
		{
			var key:String = event.currentTarget.name;
			var obj:Object = scrollZoomLayoutMouseTrack[key];
			
			//trace("MOUSE_["+event.type+"]: " + event.target.name + " - " + event.currentTarget.name);
			if (!obj)
				return;
			if (event.type == MouseEvent.MOUSE_DOWN)
			{
				obj.point = new Point(event.stageX, event.stageY);
				scrollZoomLayoutMouseTrack[key] = obj;
			}
			else if (event.type == MouseEvent.MOUSE_UP)
			{
				obj.point = null;
				scrollZoomLayoutMouseTrack[key] = obj;
			}
			else if (event.type == MouseEvent.MOUSE_MOVE)
			{
				var point:Point = obj.point;
				if (!obj.point)
					return;
				var deltaX:Number = point.x - event.stageX;
				var deltaY:Number = point.y - event.stageY;
				obj.point = new Point(event.stageX, event.stageY);
				scrollZoomLayoutMouseTrack[key] = obj;
				
				var visualContainer:IVisualElementContainer = (event.currentTarget as IVisualElementContainer);
				
				//trace("deltaX: " + deltaX + " / " + deltaY + " scrollX: " + obj.scrollX + " scrollY: " + obj.scrollY);
				
				// TODO Need to allow for Sliding to extremes with large delta's
				var left:Number = (visualContainer as UIComponent).getStyle("left");
				if (obj.scrollX > 0 && left-deltaX < 0 && left-deltaX > -obj.scrollX) 
					(visualContainer as UIComponent).setStyle("left", left-deltaX);
				var top:Number = (visualContainer as UIComponent).getStyle("top");
				if (obj.scrollY > 0 && top-deltaY < 0 && top-deltaY > -obj.scrollY) 
					(visualContainer as UIComponent).setStyle("top", top-deltaY);
				
			}
		}
		
		
		protected static function resizeComponent(element:UIComponent, newRatio:Number, filterStyles:Array):void
		{
			if (element)
			{
				element.scaleX = element.scaleY = newRatio;
				var value:Number = 0;
				var key:String = "";
				var obj:Object;
				for each (var styleName:String in filterStyles)
				{
					key = element.name+"___"+styleName;
					value = element.getStyle(styleName);
					if (!isNaN(value))
					{
						if (!originalStyleValues[key])
						{
							originalStyleValues[key] = {original:value, ratio: newRatio};
							obj = originalStyleValues[key];
						}
						else
						{
							obj = originalStyleValues[key];
							if (value / obj.ratio != obj.original)
								obj.original = value;
							if (obj.ratio != newRatio)
								obj.ratio = newRatio;
							originalStyleValues[key] = obj;
						}
						
						element.setStyle(styleName, obj.original * newRatio);
					}
				}
			}
		}
	}
}